﻿using Holo.Virtual.Rooms.Bots;
using Holo.Virtual.Rooms.Pets;
using Holo.Virtual.Users;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Holo.Virtual.Rooms.Pathfinding
{
    public class Vector2D
    {
        public int X;
        public int Y;

        public Vector2D(int x, int y)
        {
            this.X = x;
            this.Y = y;
        }

        public int GetDistanceSquared(Vector2D Point)
        {
            int dx = this.X - Point.X;
            int dy = this.Y - Point.Y;
            return (dx * dx) + (dy * dy);
        }
    }

    static class Pathinder
    {
        public static List<Vector2D> FindPath(Object roomUser, Vector2D Start, Vector2D End)
        {
            List<Vector2D> Path = new List<Vector2D>();

            PathFinderNode Nodes = FindPathReversed(roomUser, End, Start);

            if (Nodes != null) 
            {
                while (Nodes.Next != null)
                {
                    Path.Add(Nodes.Next.Position);
                    Nodes = Nodes.Next;
                }
            }

            return Path;
        }

        private static PathFinderNode FindPathReversed(Object roomUser, Vector2D Start, Vector2D End)
        {
            virtualRoom room = null;

            if (roomUser is virtualRoomUser)
            {
                room = ((virtualRoomUser)roomUser).Room;
            }

            if (roomUser is virtualBot)
            {
                room = ((virtualBot)roomUser).Room;
            }

            if (roomUser is virtualPet)
            {
                room = ((virtualPet)roomUser).Room;
            }

            if (room == null)
            {
                return null;
            }

            MinHeap<PathFinderNode> OpenList = new MinHeap<PathFinderNode>(256);

            PathFinderNode[,] Map = new PathFinderNode[room.colX, room.colY];
            PathFinderNode Node;
            Vector2D Tmp;
            int Cost;
            int Diff;

            PathFinderNode Current = new PathFinderNode(Start);
            Current.Cost = 0;

            PathFinderNode Finish = new PathFinderNode(End);
            Map[Current.Position.X, Current.Position.Y] = Current;
            OpenList.Add(Current);

            while (OpenList.Count > 0)
            {
                Current = OpenList.ExtractFirst();
                Current.InClosed = true;

                for (int i = 0; i < MovePoints.Length; i++)
                {
                    Tmp = new Vector2D(MovePoints[i].X + Current.Position.X, MovePoints[i].Y + Current.Position.Y);
                    bool IsFinalMove = (Tmp.X == End.X && Tmp.Y == End.Y); // are we at the final position?

                    if (room.isValidStep(roomUser, Current.Position.X, Current.Position.Y, Tmp.X, Tmp.Y, IsFinalMove)) // need to set the from positions
                    {
                        if (Map[Tmp.X, Tmp.Y] == null)
                        {
                            Node = new PathFinderNode(Tmp);
                            Map[Tmp.X, Tmp.Y] = Node;
                        }
                        else
                        {
                            Node = Map[Tmp.X, Tmp.Y];
                        }

                        if (!Node.InClosed)
                        {
                            Diff = 0;

                            if (Current.Position.X != Node.Position.X)
                            {
                                Diff += 1;
                            }

                            if (Current.Position.Y != Node.Position.Y)
                            {
                                Diff += 1;
                            }

                            Cost = Current.Cost + Diff + Node.Position.GetDistanceSquared(End);

                            if (Cost < Node.Cost)
                            {
                                Node.Cost = Cost;
                                Node.Next = Current;
                            }

                            if (!Node.InOpen)
                            {
                                if (Node.Equals(Finish))
                                {
                                    Node.Next = Current;
                                    return Node;
                                }

                                Node.InOpen = true;
                                OpenList.Add(Node);
                            }
                        }
                    }
                }
            }

            return null;

        }

        private static Vector2D[] MovePoints = new Vector2D[]
        {
            new Vector2D(-1, -1),
            new Vector2D(0, -1),
            new Vector2D(1, -1),
            new Vector2D(1, 0),
            new Vector2D(1, 1),
            new Vector2D(0, 1),
            new Vector2D(-1, 1),
            new Vector2D(-1, 0)
        };

        private static Vector2D[] NoDiagMovePoints = new Vector2D[]
        {
            new Vector2D(0, -1),
            new Vector2D(1, 0),
            new Vector2D(0, 1),
            new Vector2D(-1, 0)
        };
    }

    sealed class MinHeap<T> where T : IComparable<T>
    {
        private int count;
        private int capacity;
        private T temp;
        private T mheap;
        private T[] array;
        private T[] tempArray;

        public int Count
        {
            get { return this.count; }
        }

        public MinHeap() : this(16) { }

        public MinHeap(int capacity)
        {
            this.count = 0;
            this.capacity = capacity;
            array = new T[capacity];
        }

        public void BuildHead()
        {
            int position;
            for (position = (this.count - 1) >> 1; position >= 0; position--)
            {
                this.MinHeapify(position);
            }
        }

        public void Add(T item)
        {
            this.count++;
            if (this.count > this.capacity)
            {
                DoubleArray();
            }
            this.array[this.count - 1] = item;
            int position = this.count - 1;

            int parentPosition = ((position - 1) >> 1);

            while (position > 0 && array[parentPosition].CompareTo(array[position]) > 0)
            {
                temp = this.array[position];
                this.array[position] = this.array[parentPosition];
                this.array[parentPosition] = temp;
                position = parentPosition;
                parentPosition = ((position - 1) >> 1);
            }
        }

        private void DoubleArray()
        {
            this.capacity <<= 1;
            tempArray = new T[this.capacity];
            CopyArray(this.array, tempArray);
            this.array = tempArray;
        }

        private static void CopyArray(T[] source, T[] destination)
        {
            int index;
            for (index = 0; index < source.Length; index++)
            {
                destination[index] = source[index];
            }
        }

        public T Peek()
        {
            if (this.count == 0)
            {
                throw new InvalidOperationException("Heap is empty");
            }
            return this.array[0];
        }


        public T ExtractFirst()
        {
            if (this.count == 0)
            {
                throw new InvalidOperationException("Heap is empty");
            }
            temp = this.array[0];
            this.array[0] = this.array[this.count - 1];
            this.count--;
            this.MinHeapify(0);
            return temp;
        }

        private void MinHeapify(int position)
        {
            do
            {
                int left = ((position << 1) + 1);
                int right = left + 1;
                int minPosition;

                if (left < count && array[left].CompareTo(array[position]) < 0)
                {
                    minPosition = left;
                }
                else
                {
                    minPosition = position;
                }

                if (right < count && array[right].CompareTo(array[minPosition]) < 0)
                {
                    minPosition = right;
                }

                if (minPosition != position)
                {
                    mheap = this.array[position];
                    this.array[position] = this.array[minPosition];
                    this.array[minPosition] = mheap;
                    position = minPosition;
                }
                else
                {
                    return;
                }

            } while (true);
        }
    }

    sealed class PathFinderNode : IComparable<PathFinderNode>
    {
        public Vector2D Position;
        public PathFinderNode Next;
        public int Cost = Int32.MaxValue;
        public bool InOpen = false;
        public bool InClosed = false;

        public PathFinderNode(Vector2D Position)
        {
            this.Position = Position;
        }

        public override bool Equals(object obj)
        {
            return (obj is PathFinderNode) && ((PathFinderNode)obj).Position.Equals(this.Position);
        }

        public bool Equals(PathFinderNode Breadcrumb)
        {
            if (Breadcrumb.Position.X == this.Position.X && Breadcrumb.Position.Y == this.Position.Y)
            {
                return true;
            }

            return false;
        }

        public override int GetHashCode()
        {
            return Position.GetHashCode();
        }

        public int CompareTo(PathFinderNode Other)
        {
            return Cost.CompareTo(Other.Cost);
        }
    }
}
